from datetime import datetime
from decimal import Decimal


class Field:
    SQL_TYPE = None

    def __init__(self, pk=False, null=False, unique=False, default=None, column_name=None):
        self._pk = pk
        self._null = null
        self._unique = unique
        self._default = default
        self._column_name = column_name
        self._name = None
        self._model = None

    def adapt(self, value):
        return value

    def adapt_value(self, value):
        if value is None and self._null:
            return None

        return self.adapt(value)

    def to_query_parameter(self, value):
        return value

    def render_sql(self):
        column_name = self.column_name
        sql_type = self.render_sql_type()

        constrains = self.get_field_constrains()

        sql_parts = [column_name, sql_type, *constrains]
        return ' '.join(sql_parts)

    def get_field_constrains(self):
        constrains = []
        if self._pk:
            constrains.append('PRIMARY KEY')
        if not self._null:
            constrains.append('NOT NULL')
        if self._unique:
            constrains.append('UNIQUE')
        return constrains

    def render_sql_type(self):
        return self.SQL_TYPE

    def get_default(self):
        if callable(self._default):
            return self._default()
        return self._default

    @property
    def column_name(self):
        return self._column_name

    @property
    def name(self):
        return self._name

    @property
    def model(self):
        return self._model

    def __set_name__(self, owner, name):
        self._model = owner
        self._name = name
        if not self._column_name:
            self._column_name = name

    @property
    def is_pk(self):
        return bool(self._pk)

    @property
    def query_name(self):
        return f'{self.model._meta.table_name}.{self.column_name}'


class IntegerField(Field):
    SQL_TYPE = 'INTEGER'
    adapt = int


class BooleanField(Field):
    SQL_TYPE = 'BOOLEAN'
    adapt = bool


class CharField(Field):

    def __init__(self, pk=False, null=False, unique=False, default=None, column_name=None, **extra_kwargs):
        max_length = extra_kwargs.pop('max_length')
        super().__init__(pk=pk, null=null, unique=unique, default=default, column_name=column_name)
        self.max_length = min(int(max_length), 255)

    def render_sql_type(self):
        return f'VARCHAR({self.max_length})'

    def adapt(self, value):
        if not value:
            return ''
        if isinstance(value, str):
            return value
        if isinstance(value, bytes):
            return value.decode('utf-8')
        return str(value)


class DecimalField(Field):

    def __init__(self, pk=False, null=False, unique=False, default=None, column_name=None, **extra_kwargs):
        max_digits = extra_kwargs.pop('max_digits')
        decimal_places = extra_kwargs.pop('decimal_places')

        super().__init__(pk, null=null, unique=unique, default=default, column_name=column_name)

        self._max_digits = max_digits
        self._decimal_places = decimal_places

    def render_sql_type(self):
        return f'DECIMAL({self._max_digits}, {self._decimal_places})'

    def adapt(self, value):
        if not isinstance(value, Decimal):
            if isinstance(value, float):
                dec = self._decimal_places
                return Decimal(round(value * (10 ** dec))) / Decimal(10 ** dec)
            return Decimal(value)
        return value


class DateTimeField(Field):
    SQL_TYPE = 'DATETIME'
    FORMATS = ('%Y-%m-%d %H:%M:%S.%f', '%Y-%m-%d %H:%M:%S', '%Y-%m-%d')

    def adapt(self, value):
        if isinstance(value, str):
            for fmt in self.FORMATS:
                try:
                    return datetime.strptime(value, fmt)
                except ValueError:
                    pass
        return value


class AutoField(Field):

    def __init__(self, pk=False, null=False, unique=False, default=None, column_name=None):
        super(AutoField, self).__init__(pk=pk, null=null, unique=unique, default=default, column_name=column_name)
        self._null = self.is_pk
        self._unique = False

    def render_sql_type(self):
        field_type = self.model._meta.db.spec.auto_field_type
        return field_type

    def get_field_constrains(self):
        constrains = super().get_field_constrains()
        constrains.extend(self.model._meta.db.spec.auto_field_constrains)
        return constrains


class ForeignKey(Field):

    def __init__(self, to, pk=False, null=False, unique=False, default=None, column_name=None):
        if not column_name:
            column_name = f"{to._meta.table_name}_id"

        super().__init__(pk=pk, null=null, unique=unique, default=default, column_name=column_name)
        self.to = to

    def adapt(self, value):
        if isinstance(value, self.to):
            return value.pk
        return self.to._meta.pk_field.adapt(value)

    def to_query_parameter(self, value):
        return value.pk if isinstance(value, self.to) else value

    def render_sql_type(self):
        ref_pk_field = self.to._meta.pk_field
        fk_type = 'INTEGER' if isinstance(ref_pk_field, AutoField) else ref_pk_field.render_sql_type()
        return f'{fk_type} REFERENCES {self.to._meta.table_name} ({ref_pk_field.name})'



from collections import namedtuple

from minorm.connectors import connector
from minorm.exceptions import DoesNotExists
from minorm.expressions import WhereCondition
from minorm.fields import AutoField, Field, ForeignKey
from minorm.managers import QuerySet
from minorm.queries import CreateTableQuery, DeleteQuery, DropTableQuery, InsertQuery, UpdateQuery, SelectQuery


class ModelSetupError(Exception):
    pass


class ModelMetaData:

    def __init__(self, model_name, db, table_name, fields):
        self._model_name = model_name
        self._db = db
        self._table_name = table_name
        self._fields = fields

    @property
    def db(self):
        return self._db

    @property
    def table_name(self):
        return self._table_name

    @property
    def fields(self):
        return list(self._fields)

    @property
    def pk_field(self):
        return next((field for field in self.fields if field.is_pk))

    @property
    def name(self):
        return self._model_name.lower()

    @property
    def column_names(self):
        return [field.column_name for field in self.fields]

    @property
    def query_names(self):
        return [field.query_name for field in self.fields]

    def check_field(self, field_name, with_pk=False):
        for field in self.fields:
            if field.name == field_name and (not field.is_pk or with_pk):
                return field

        raise ValueError(f'{field_name} is not a valid field for model {self._model_name}.')

    def get_field(self, field_name):
        return self.check_field(field_name, with_pk=True)

    def get_fk_field(self, field_name):
        for field in self.fields:
            if field.name == field_name and isinstance(field, ForeignKey):
                return field

        raise ValueError(f'{field_name} is not a valid foreign relation for model {self._model_name}.')


class ModelMetaclass(type):

    def __new__(mcs, name, bases, namespace):
        if not bases:
            return super().__new__(mcs, name, bases, namespace)

        fields = [class_attr for class_attr in namespace.values() if isinstance(class_attr, Field)]
        # Extract model meta
        meta = namespace.pop('Meta', None)
        table_name = getattr(meta, 'table_name', name.lower())
        db = getattr(meta, 'db', connector)

        queryset_class = namespace.pop('queryset_class', QuerySet)

        # Extract primary key:
        pk_fields = [field for field in fields if field.is_pk]
        if len(pk_fields) > 1:
            raise ModelSetupError('Model should have only one primary key.')
        if pk_fields:
            pk_field = pk_fields[0]
        else:
            pk_field = AutoField(pk=True, column_name='id')
            setattr(pk_field, '_name', 'id')
        fields.append(pk_field)

        # Setup model pk, meta and queryset attributes:
        model = super().__new__(mcs, name, bases, namespace)

        if not pk_field.model:
            setattr(pk_field, '_model', model)

        setattr(model, '_meta', ModelMetaData(model_name=model.__name__, db=db, table_name=table_name, fields=fields))

        setattr(model, '_queryset_class', queryset_class)

        does_not_exists = type(f'{model.__name__}DoesNotExists', (DoesNotExists,), {})
        setattr(model, 'DoesNotExists', does_not_exists)

        query_namedtuple = namedtuple(f'{model.__name__}QueryNamedTuple', field_names=[field.name for field in fields])
        setattr(model, 'query_namedtuple', query_namedtuple)

        return model

    @property
    def qs(cls):
        return cls._queryset_class(model=cls)

    def render_sql(cls):
        field_params = [field.render_sql() for field in cls._meta.fields]
        create_query = CreateTableQuery(table_name=cls._meta.table_name, params=field_params)
        return create_query.render_sql()

    def create_table(cls):
        raw_sql = cls.render_sql()
        with cls._meta.db.cursor() as curr:
            curr.execute(raw_sql)

    def drop_table(cls):
        drop_query = DropTableQuery(table_name=cls._meta.table_name)
        raw_sql = drop_query.render_sql()
        with cls._meta.db.cursor() as curr:
            curr.execute(raw_sql)


class Model(metaclass=ModelMetaclass):

    def __init__(self, **kwargs):
        for field in self.__class__._meta.fields:
            field_name = field.name
            value = kwargs.get(field_name, field.get_default())
            setattr(self, field_name, value)

    @property
    def pk(self):
        return getattr(self, self.__class__._meta.pk_field.name)

    def save(self):
        is_creation = not bool(self.pk)
        self._adapt_values()

        model = self.__class__

        modified_fields = [field for field in model._meta.fields if not isinstance(field, AutoField)]
        if is_creation:
            query_class = InsertQuery
            where_cond = None
        else:
            query_class = UpdateQuery
            where_cond = WhereCondition(model._meta.pk_field.query_name, WhereCondition.EQ, self.pk)

        query = query_class(
            table_name=model._meta.table_name,
            fields=[field.column_name for field in modified_fields],
            where=where_cond,
        )
        raw_sql = query.render_sql(model._meta.db.spec)
        query_params = [getattr(self, field.name) for field in modified_fields]
        if where_cond:
            query_params.extend(where_cond.values())
        with model._meta.db.cursor() as curr:
            curr.execute(raw_sql, query_params)
        if is_creation:
            setattr(self, model._meta.pk_field.name, curr.lastrowid)

    def _adapt_values(self):
        for field in self.__class__._meta.fields:
            field_name = field.name
            adapted_value = field.adapt_value(getattr(self, field_name))
            setattr(self, field_name, adapted_value)

    def refresh_from_db(self):
        if not self.pk:
            return

        model = self.__class__

        pk_cond = WhereCondition(model._meta.pk_field.query_name, WhereCondition.EQ, self.pk)
        query_fields = [field.query_name for field in model._meta.fields]
        select_query = SelectQuery(table_name=model._meta.table_name, fields=query_fields, where=pk_cond)
        raw_sql = select_query.render_sql(model._meta.db.spec)
        params = pk_cond.values()
        with model._meta.db.cursor() as curr:
            curr.execute(raw_sql, params)
            row = curr.fetchone()
        for i, field in enumerate(model._meta.fields):
            setattr(self, field.name, row[i])

    def delete(self):
        if not self.pk:
            return

        model = self.__class__
        pk_cond = WhereCondition(model._meta.pk_field.query_name, WhereCondition.EQ, self.pk)
        delete_query = DeleteQuery(table_name=model._meta.table_name, where=pk_cond)

        raw_sql = delete_query.render_sql(model._meta.db.spec)
        with model._meta.db.cursor() as curr:
            curr.execute(raw_sql, pk_cond.values())

        setattr(self, model._meta.pk_field.name, None)
        return curr.rowcount



from collections import namedtuple


LOOKUP_SEPARATOR = '__'


class WhereCondition:
    AND = 'AND'
    OR = 'OR'
    NOT = 'NOT'

    EQ = '='

    IN = 'IN'
    LIKE = 'LIKE'

    LOOKUP_MAPPING = (
        ('lt', '<'),
        ('lte', '<='),
        ('gt', '>'),
        ('gte', '>='),
        ('in', IN),
        ('neq', '!='),
        ('startswith', LIKE),
        ('endswith', LIKE),
        ('contains', LIKE),
    )

    LIKE_PATTERNS = (
        ('startswith', '{0}%'),
        ('endswith', '%{0}'),
        ('contains', '%{0}%')
    )

    MULTIPLE_VALUE_OPS = (IN, )

    def __init__(self, field, op, value, no_escape=False):
        self.field = field
        self.op = op
        self.value = value

        self.no_escape = no_escape

        self._and = None
        self._or = None

        self._negated = False

    def __str__(self):
        result = f'{self.field} {self.op} {self.value if self.no_escape else self.resolved_escape}'

        if self._and:
            result = f'{result} {self.AND} {self._and}'

        if self._or:
            result = f'{result} {self.OR} {self._or}'

        if self._negated:
            result = f'{self.NOT} ({result})'

        return result

    def values(self):
        value = self.value
        result = tuple(value) if self.op in self.MULTIPLE_VALUE_OPS else (value, )

        if self._and:
            result += self._and.values()

        if self._or:
            result += self._or.values()

        return result

    @property
    def resolved_escape(self):
        if self.op in self.MULTIPLE_VALUE_OPS:
            return f"({', '.join('{0}' for _ in range(len(self.value)))})"
        return '{0}'

    def __and__(self, other):
        self._and = other
        return self

    def __or__(self, other):
        self._or = other
        return self

    def __invert__(self):
        self._negated = not self._negated
        return self

    @classmethod
    def resolve_lookup(cls, lookup_key):
        *rest, lookup = lookup_key.split(LOOKUP_SEPARATOR)

        lookups = dict(cls.LOOKUP_MAPPING)
        if lookup not in lookups:
            lookup = None
            field_name = lookup_key
        else:
            field_name = LOOKUP_SEPARATOR.join(rest)
        return field_name, lookup

    @classmethod
    def for_lookup(cls, field_name, lookup, value):
        if lookup is None:
            op = cls.EQ
        else:
            lookup_mapping = dict(cls.LOOKUP_MAPPING)
            op = lookup_mapping[lookup]

        if op == cls.LIKE:
            like_patterns = dict(cls.LIKE_PATTERNS)
            value = like_patterns[lookup].format(value)

        return cls(field=field_name, op=op, value=value)


class OrderByExpression(namedtuple('OrderByExpression', 'value, ordering')):
    ASC = 'ASC'
    DESC = 'DESC'

    @classmethod
    def from_field_name(cls, field_name):
        if field_name.startswith('-'):
            value = field_name[1:]
            ordering = cls.DESC
        else:
            value = field_name
            ordering = cls.ASC

        return cls(value=value, ordering=ordering)

    def __str__(self):
        return f'{self.value} {self.ordering}'


class JoinExpression:
    LEFT_OUTER = 'LEFT OUTER'
    INNER = 'INNER'

    def __init__(self, table_name, on, join_type=LEFT_OUTER):
        self.table_name = table_name
        self.on = on
        self.join_type = join_type

    def __str__(self):
        return f'{self.join_type} JOIN {self.table_name} ON {self.on}'

    @classmethod
    def on_pk(cls, outer_table, pk_field, fk_field, join_type=None):
        on = WhereCondition(pk_field, WhereCondition.EQ, fk_field, no_escape=True)
        return cls(table_name=outer_table, on=on, join_type=(join_type or cls.LEFT_OUTER))





class DDLQuery:

    def __init__(self, table_name, params=()):
        self.table_name = table_name
        self.params = params

    def render_sql(self):
        raise NotImplementedError


class CreateTableQuery(DDLQuery):
    TEMPLATE = 'CREATE TABLE {table} ({fields});'

    def render_sql(self):
        field_part = ', '.join(self.params)
        return self.TEMPLATE.format(table=self.table_name, fields=field_part)


class DropTableQuery(DDLQuery):
    TEMPLATE = 'DROP TABLE {table};'

    def render_sql(self):
        return self.TEMPLATE.format(table=self.table_name)


class DMLQuery:

    def __init__(self, table_name, fields=(), where=None, limit=None):
        self.table_name = table_name
        self.fields = fields

        self._where = where
        self._limit = limit

    def where(self, expr):
        self._where = expr
        return self

    def limit(self, value):
        self._limit = value
        return self

    def render_sql(self, db_spec):
        raise NotImplementedError


class InsertQuery(DMLQuery):

    def render_sql(self, db_spec):
        fields_part = ', '.join(self.fields)

        value_escape = db_spec.value_escape
        values_part = ', '.join(value_escape for _ in self.fields)

        result = f'INSERT INTO {self.table_name} ({fields_part}) VALUES ({values_part});'
        return result


class UpdateQuery(DMLQuery):

    def render_sql(self, db_spec):
        value_escape = db_spec.value_escape
        fields_part = ', '.join(f'{field} = {value_escape}' for field in self.fields)

        update_str = f'UPDATE {self.table_name} SET {fields_part}'
        query_parts = [update_str]

        if self._where:
            where_part = f'WHERE {self._where}'
            where_part_proper_escape = where_part.format(value_escape)
            query_parts.append(where_part_proper_escape)

        return f"{' '.join(query_parts)};"


class DeleteQuery(DMLQuery):

    def render_sql(self, db_spec):
        update_str = f'DELETE FROM {self.table_name}'
        query_parts = [update_str]

        if self._where:
            value_escape = db_spec.value_escape
            where_part = f'WHERE {self._where}'
            where_part_proper_escape = where_part.format(value_escape)
            query_parts.append(where_part_proper_escape)

        return f"{' '.join(query_parts)};"


class SelectQuery(DMLQuery):

    def __init__(self, table_name, fields=(), where=None):
        super().__init__(table_name, fields, where)

        self._joins = []
        self._order_by = None

    def render_sql(self, db_spec):
        fields_part = ', '.join(self.fields)

        select_str = f'SELECT {fields_part} FROM {self.table_name}'
        query_parts = [select_str]

        query_parts.extend(str(join) for join in self._joins)

        if self._where:
            value_escape = db_spec.value_escape

            where_part = f'WHERE {self._where}'
            where_part_proper_escape = where_part.format(value_escape)
            query_parts.append(where_part_proper_escape)

        if self._order_by:
            order_part = ', '.join(str(ordering) for ordering in self._order_by)
            order_str = f'ORDER BY {order_part}'
            query_parts.append(order_str)

        if self._limit:
            limit_str = f'LIMIT {self._limit}'
            query_parts.append(limit_str)

        return f"{' '.join(query_parts)};"

    def join(self, join_expression):
        self._joins.extend(join_expression)
        return self

    def order_by(self, order_expression):
        self._order_by = order_expression
        return self



from decimal import Decimal


class BaseSpec:
    VALUE_ESCAPE = None
    AUTO_FIELD_TYPE = None
    AUTO_FIELD_CONSTRAINS = ()

    def __init__(self, connection_url):
        assert self.VALUE_ESCAPE, f"{self.__class__.__name__} should define value escape."
        assert self.AUTO_FIELD_TYPE,  f"{self.__class__.__name__} should define auto field type."

        self.connection_url = connection_url  # TODO: add validation for certain database

        self.db_driver = self.prepare_db_driver()

    def prepare_db_driver(self):
        raise NotImplementedError

    def create_connection(self):
        connection = self.db_driver.connect(self.connection_url)
        return connection

    @property
    def value_escape(self):
        return str(self.VALUE_ESCAPE)

    @property
    def auto_field_type(self):
        return str(self.AUTO_FIELD_TYPE)

    @property
    def auto_field_constrains(self):
        return tuple(self.AUTO_FIELD_CONSTRAINS)


class SQLiteSpec(BaseSpec):
    VALUE_ESCAPE = '?'
    AUTO_FIELD_TYPE = "INTEGER"
    AUTO_FIELD_CONSTRAINS = ("AUTOINCREMENT",)

    def prepare_db_driver(self):
        import sqlite3

        sqlite3.register_adapter(Decimal, str)
        sqlite3.register_converter("DECIMAL", Decimal)

        return sqlite3


class PostgreSQLSpec(BaseSpec):
    VALUE_ESCAPE = '%s'
    AUTO_FIELD_TYPE = "SERIAL"
    HAS_AUTO_FIELD_AUTO_INCREMENT = False

    def prepare_db_driver(self):
        import psycopg2
        return psycopg2



from collections import OrderedDict
import functools
import operator

from minorm.exceptions import MultipleQueryResult
from minorm.expressions import JoinExpression, LOOKUP_SEPARATOR, OrderByExpression, WhereCondition
from minorm.queries import DeleteQuery, InsertQuery, SelectQuery, UpdateQuery


class RelationNode:
    """A helper class for constructing nested foreign relations."""

    def __init__(self, base_model, depth=0, position=0):
        self.model = base_model

        self.depth = depth
        self.position = position

        self.is_selected = False  # should be True when the relation is marked in `select_related` method

        self.relations = OrderedDict()

    def resolve_relation(self, lookup_parts, is_selected=False):
        field_name, *rest_lookup_parts = lookup_parts
        if field_name not in self.relations:
            field = self.model._meta.get_fk_field(field_name)
            position = len(self.relations) + 1
            self.relations[field_name] = RelationNode(base_model=field.to, depth=self.depth + 1, position=position)

        if is_selected:
            self.relations[field_name].is_selected = True

        if rest_lookup_parts:
            return self.relations[field_name].resolve_relation(rest_lookup_parts, is_selected=is_selected)
        return self.relations[field_name]

    @property
    def is_root_node(self):
        return self.depth == self.position == 0

    @property
    def table_shortcut(self):
        if self.is_root_node:
            return self.model._meta.table_name

        return f'T{self.depth}{self.position}'

    @property
    def table_name(self):
        if self.is_root_node:
            return self.model._meta.table_name

        return f'{self.model._meta.table_name} {self.table_shortcut}'

    def get_column_names(self):
        column_names = [f'{self.table_shortcut}.{field.column_name}' for field in self.model._meta.fields]
        for rel in self.relations.values():
            if rel.is_selected:
                column_names.extend(rel.get_column_names())
        return column_names

    def get_joins(self):
        joins = []
        for attr_name, rel in self.relations.items():
            fk = self.model._meta.get_fk_field(attr_name)
            joins.append(
                JoinExpression.on_pk(
                    outer_table=rel.table_name,
                    fk_field=f'{self.table_shortcut}.{fk.column_name}',
                    pk_field=f'{rel.table_shortcut}.{fk.to._meta.pk_field.column_name}',
                )
            )
            joins.extend(rel.get_joins())

        return joins

    def row_to_instance(self, row, is_namedtuple, row_shift=0):
        model = self.model

        row_part = row[row_shift:]
        attr_names = (field.name for field in model._meta.fields)
        kwargs = dict(zip(attr_names, row_part))
        row_shift += len(kwargs)

        for fk_name, rel in self.relations.items():
            if rel.is_selected:
                kwargs[fk_name], row_shift = rel.row_to_instance(row, is_namedtuple, row_shift=row_shift)

        klass = model.query_namedtuple if is_namedtuple else model
        return klass(**kwargs), row_shift


class QuerySet:

    def __init__(self, model):
        self.model = model

        self._where = None
        self._order_by = set()
        self._limit = None

        self._related = RelationNode(base_model=self.model)
        self._values_mapping = OrderedDict()

    def filter(self, **kwargs):
        where_cond = self._where_action(**kwargs)
        self._reset_where(where_cond, operator.and_)
        return self

    def aswell(self, **kwargs):
        where_cond = self._where_action(**kwargs)
        self._reset_where(where_cond, operator.or_)
        return self

    def order_by(self, *args):
        for field_name in args:
            order_exp = OrderByExpression.from_field_name(field_name)
            field = self.model._meta.check_field(order_exp.value, with_pk=True)
            self._order_by.add(OrderByExpression(value=field.query_name, ordering=order_exp.ordering))
        return self

    def select_related(self, *args):
        if len(args) == 1 and args[0] is None:
            self._related = RelationNode(base_model=self.model)
            return self

        for lookup in args:
            lookup_parts = lookup.split(LOOKUP_SEPARATOR)
            self._related.resolve_relation(lookup_parts, is_selected=True)

        return self

    def update(self, **kwargs):
        update_data = OrderedDict()
        for key, value in kwargs.items():
            field = self.model._meta.check_field(key, with_pk=False)
            adopted_value = field.adapt_value(value)
            update_data[field.column_name] = adopted_value

        update_query = UpdateQuery(
            table_name=self.model._meta.table_name,
            fields=update_data.keys(),
            where=self._where,
        )
        raw_sql = update_query.render_sql(self.model._meta.db.spec)
        params = tuple(update_data.values()) + (self._where.values() if self._where else ())
        with self.model._meta.db.cursor() as curr:
            curr.execute(raw_sql, params)
        return curr.rowcount

    def delete(self):
        delete_query = DeleteQuery(table_name=self.model._meta.table_name, where=self._where)
        raw_sql = delete_query.render_sql(self.model._meta.db.spec)

        with self.model._meta.db.cursor() as curr:
            curr.execute(raw_sql, self.query_params)
        return curr.rowcount

    def create(self, **kwargs):
        instance = self.model(**kwargs)
        instance.save()
        return instance

    def get(self, **kwargs):
        self._values_mapping = {}
        self._limit = 2
        results = self._extract(**kwargs)
        if not results:
            raise self.model.DoesNotExists
        if len(results) > 1:
            raise MultipleQueryResult

        return self._instance_from_row(results[0])

    def first(self):
        self._limit = 1
        results = self._extract()
        if not results:
            return None

        if self._values_mapping:
            return self._dict_from_row(results[0])
        return self._instance_from_row(results[0])

    def all(self):
        extracted = self._extract()
        if self._values_mapping:
            return [self._dict_from_row(row) for row in extracted]
        return [self._instance_from_row(row, is_namedtuple=True) for row in extracted]

    def values(self, *args):
        if len(args) == 1 and args[0] is None:
            self._values_mapping = OrderedDict()
            return self

        for lookup in args:
            val_parts = lookup.split(LOOKUP_SEPARATOR)
            *relation_lookup, field_name = val_parts
            if relation_lookup:
                relation = self._related.resolve_relation(relation_lookup)
            else:
                relation = self._related

            field = relation.model._meta.check_field(field_name, with_pk=True)
            self._values_mapping[lookup] = f'{relation.table_shortcut}.{field.column_name}'

        return self

    def exists(self):
        self.select_related(None)
        self.values(self.model._meta.pk_field.column_name)
        self._limit = 1

        is_exists = bool(self._extract())
        return is_exists

    def bulk_create(self, instances):
        model = self.model
        fields = model._meta.fields
        db = model._meta.db

        insert_query = InsertQuery(table_name=model._meta.table_name, fields=[field.column_name for field in fields])
        raw_sql = insert_query.render_sql(db.spec)
        params = [[getattr(obj, field.name) for field in fields] for obj in instances if isinstance(obj, model)]
        with db.cursor() as curr:
            curr.executemany(raw_sql, params)
        return curr.rowcount

    @property
    def query(self):
        column_names = self._values_mapping.values() if self._values_mapping else self._related.get_column_names()
        joins = self._related.get_joins()

        query = (SelectQuery(table_name=self._related.table_name, fields=column_names)
                 .join(joins)
                 .where(self._where)
                 .limit(self._limit)
                 .order_by(self._order_by))
        return query

    @property
    def query_params(self):
        return self._where.values() if self._where else ()

    def __getitem__(self, item):
        if not any(isinstance(item, supported_type) for supported_type in (int, slice)):
            raise TypeError(f'{self.__class__.__name__} indices must be integers or slices.')

        if isinstance(item, slice):
            self._limit = item.stop
            return self

        return self.all()[item]

    def _where_action(self, *args, **kwargs):
        where_conds = self._check_pk_lookups(kwargs)
        where_conds.extend(args)

        for key, value in kwargs.items():
            field_part, lookup = WhereCondition.resolve_lookup(key)
            *rel_lookups, field_name = field_part.split(LOOKUP_SEPARATOR)
            if rel_lookups:
                related = self._related.resolve_relation(rel_lookups)
            else:
                related = self._related

            field = related.model._meta.check_field(field_name, with_pk=True)
            adopted_value = field.to_query_parameter(value)
            where_cond = WhereCondition.for_lookup(
                f'{related.table_shortcut}.{field.column_name}', lookup, adopted_value
            )
            where_conds.append(where_cond)

        result = functools.reduce(operator.and_, where_conds) if where_conds else None
        return result

    def _reset_where(self, where_cond, op):
        if not where_cond:
            return

        if not self._where:
            self._where = where_cond
        else:
            self._where = op(self._where, where_cond)

    def _extract(self, **kwargs):
        where_cond = self._where_action(**kwargs)
        self._reset_where(where_cond, operator.and_)

        select_query = self.query
        params = self.query_params

        raw_sql = select_query.render_sql(self.model._meta.db.spec)

        with self.model._meta.db.cursor() as curr:
            curr.execute(raw_sql, params)
            results = curr.fetchall()
        return results

    def _check_pk_lookups(self, kwargs):
        conds = []
        lookups = set()

        pk_prefix = f'{self.model._meta.pk_field.name}__'
        for key, value in kwargs.items():
            if not key.startswith(pk_prefix):
                continue

            _, lookup = WhereCondition.resolve_lookup(key)
            where_cond = WhereCondition.for_lookup(self.model._meta.pk_field.query_name, lookup, value)
            conds.append(where_cond)
            lookups.add(key)

        for lookup in lookups:
            del kwargs[lookup]

        return conds

    def _instance_from_row(self, row, is_namedtuple=False):
        instance, __ = self._related.row_to_instance(row, is_namedtuple=is_namedtuple)
        return instance

    def _dict_from_row(self, row):
        return dict(zip(self._values_mapping, row))



from pathlib import Path
from setuptools import setup


def get_version(root_path):
    version_file = root_path / 'minorm' / '__init__.py'
    with version_file.open() as f:
        for line in f:
            if line.startswith('__version__'):
                return line.split('=')[1].strip().strip('"').strip("'")


ROOT_PATH = Path(__file__).parent
README = ROOT_PATH / 'README.rst'

setup(
    name='minorm',
    version=get_version(ROOT_PATH),
    description='A minimalistic ORM with basic features.',
    long_description=README.read_text(),
    long_description_content_type='text/x-rst',
    url='https://github.com/EliasCampos/minorm',
    author='Campos Ilya',
    author_email='camposylia@gmail.com',
    license="MIT",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    packages=['minorm'],
)



from contextlib import contextmanager


class ConnectorError(RuntimeError):
    pass


class Connector:
    NOT_CONNECTED_ERROR = 'Connect was not performed.'

    def __init__(self):
        self._connection = None
        self._db_spec = None

    def connect(self, db_spec):
        self.disconnect()

        self._db_spec = db_spec
        self._connection = self.spec.create_connection()

        return self

    def disconnect(self):
        if self._connection:
            self._connection.close()
            self._connection = None
        self._db_spec = None

    @property
    def spec(self):
        if not self._db_spec:
            raise ConnectorError(self.NOT_CONNECTED_ERROR)

        return self._db_spec

    @contextmanager
    def cursor(self):
        if not self._connection:
            raise ConnectorError(self.NOT_CONNECTED_ERROR)

        with self._connection:
            yield self._connection.cursor()


connector = Connector()
